-- Sliding mazes puzzle
--
-- Maze design taken from: https://www.youtube.com/watch?v=pHOmjbH6zdM
-- Code: copyright © 2019 Pedro Gimeno Fortea.
--
-- You can do whatever you want with this software, under the sole condition
-- that this notice and any copyright notices are preserved. It is offered
-- with no warrany, not even implied.


love.graphics.setFont(love.graphics.newFont(80))

local portland = require 'portland'

local pieces = {
  {
    map = {
      1,1,1,1,1,1,1,
      1,0,0,0,0,0,1,
      1,0,1,1,1,0,1,
      1,0,0,0,1,0,1,
      1,1,1,0,1,0,1,
      1,0,0,0,1,0,1,
      1,1,1,1,1,1,1,
      1,0,0,0,1,0,1,
      1,1,1,0,1,0,1,
      1,0,0,0,1,0,1,
      1,0,1,0,1,0,1,
      1,0,1,0,0,0,1,
      1,1,1,1,1,1,1,
      1,0,1,0,0,0,1,
      1,0,1,0,1,0,1,
      1,0,0,0,1,0,1,
      1,0,1,0,1,0,1,
      1,0,1,0,1,0,1,
      1,1,1,1,1,1,1,
      1,1,1,1,1,1,1,
      1,1,1,1,1,1,1,
      1,1,1,1,1,1,1,
      1,1,1,1,1,1,1,
      1,1,1,1,1,1,1,
    }
  },
  {
    map = {
      1,1,1,1,1,1,1,
      1,0,1,0,0,0,1,
      1,0,1,0,1,0,1,
      1,0,0,0,1,0,1,
      1,0,1,1,1,0,1,
      1,0,0,0,1,0,1,
      1,1,1,1,1,0,1,
      1,0,0,0,0,0,1,
      1,0,1,1,1,1,1,
      1,0,0,0,0,0,1,
      1,1,1,1,1,0,1,
      1,0,0,0,0,0,1,
      1,0,1,1,1,1,1,
      1,0,1,0,0,0,1,
      1,0,1,0,1,0,1,
      1,0,0,0,1,0,1,
      1,1,1,1,1,0,1,
      1,0,0,0,0,0,1,
      1,0,1,1,1,1,1,
      1,0,1,1,1,1,1,
      1,0,1,1,1,1,1,
      1,0,1,1,1,1,1,
      1,0,1,1,1,1,1,
      1,0,1,1,1,1,1,
    }
  }
}
local w = 7
local h = #pieces[1].map / w
assert(h % 1 == 0)

pieces[1].x = -w
pieces[2].x = 0
pieces[1].y = -4
pieces[2].y = -4

local sliders = {
  {
    x = -w + 1,
    y = 1,
  },
  {
    x = -w + 1,
    y = 7,
  },
  {
    x = -w + 1,
    y = 13,
  },
}


local piece = 1
local slider = 0

local radius = 1
local radiusGrab2 = 2^2

local grabmode = false
local grab_x = 0
local grab_y = 0

-- We work in our own "virtual screen".
-- It has a width of 1 + 2*w + 1
-- and a height of 1 + 2*h + 1 + 4
-- with 0, 0 at the centre
-- and uses map units.
local areaW = 1 + 2*w + 1
local areaH = 1 + 2*h + 1 + 4
local originX, originY, scale

local function setTransform()
  love.graphics.origin()
  portland.orient()
  originX = portland.width * 0.5
  originY = portland.height * 0.5
  scale = math.min(portland.width / areaW, portland.height / areaH)
end


portland.setTween(portland.ease.sigmoid5, 0.2)

setTransform()


local function scr2map(x, y)
  x, y = love.graphics.inverseTransformPoint(x, y)
  x, y = (x - originX) / scale, (y - originY) / scale
  return x, y
end

local function map2scr(x, y)
  return --love.graphics.transformPoint(
    x * scale + originX, y * scale + originY
  --)
end

local function dist2(x1, y1, x2, y2)
  return (x2 - x1)^2 + (y2 - y1)^2
end

local function sgn(x)
  return math.max(-1, math.min(1, x*1e200*1e200))
end


local won = false

function love.update(dt)
  portland.updateOrientation(dt)
  if not won and pieces[2].y <= -24 then
    won = true
  end
end


local effangle = 0

function love.draw()
  setTransform()
  local rect = love.graphics.rectangle
  local circle = love.graphics.circle
  local setColor = love.graphics.setColor

  -- Draw pieces
  for i = 1, 2 do
    if i == piece then
      setColor(1, 0.9, 0.9)
    else
      setColor(0.7, 0.7, 0.7)
    end
    for iy = 0, h - 1 do
      for ix = 0, w - 1 do
        if pieces[i].map[iy * w + ix + 1] == 1 then
          local x, y = map2scr(pieces[i].x + ix, pieces[i].y + iy)
          rect("fill", x, y, scale, scale)
        end
      end
    end
  end

  setColor(1,1,1)
  for i = 1, #sliders do
    local x, y = map2scr(sliders[i].x + 0.5, sliders[i].y + 0.5)
    circle("fill", x, y, radius * scale)
    x, y = map2scr(sliders[i].x + w + 0.5, sliders[i].y + 0.5)
    circle("fill", x, y, radius * scale)
  end
  love.graphics.setLineWidth(2)
  for i = 1, #sliders do
    if i == slider then
      setColor(1,0.5,0.5)
    else
      setColor(0.5,0.5,0.5)
    end

    local x, y = map2scr(sliders[i].x + 0.5, sliders[i].y + 0.5)
    circle("line", x, y, radius * scale)
    x, y = map2scr(sliders[i].x + w + 0.5, sliders[i].y + 0.5)
    circle("line", x, y, radius * scale)
  end

  if won then
    love.graphics.push()
    love.graphics.translate(map2scr(0, 0))
    love.graphics.rotate(math.rad(-60))
--    love.graphics.translate(-love.graphics.getFont():getHeight()*0.5, -love.graphics.getFont():getWidth("SOLVED")*0.5)
    love.graphics.setColor(1,0,1, (math.sin(effangle)+1)*0.4+0.1)
    effangle = effangle + 0.05
    love.graphics.print("SOLVED", -love.graphics.getFont():getWidth("SOLVED")*0.5,  -love.graphics.getFont():getHeight()*0.5)
    love.graphics.rectangle("fill", -3, -3, 6, 6)
    love.graphics.pop()
  end

end


function love.keypressed(k)
  if k == "escape" then return love.event.quit() end
end


function love.mousepressed(x, y, b)
  if won then return end
  if b ~= 1 then return end
  assert(not grabmode, "Touch/click while grabbed!")
  x, y = scr2map(x, y)

  -- Don't change orientation while clicked/touched!
  portland.forceOrient(portland.orientation)

  -- Sliders first
  for i = 1, 3 do
    if   dist2(sliders[i].x + 0.5, sliders[i].y + 0.5, x, y) <= radiusGrab2
      or dist2(sliders[i].x + 0.5 + w, sliders[i].y + 0.5, x, y) <= radiusGrab2
    then
      grabmode = "slider"
      grab_x = x
      grab_y = y
      slider = i
      return
    end
  end

  for i = 1, 2 do
    local xg, yg = math.floor(x - pieces[i].x), math.floor(y - pieces[i].y)
    if xg >= 0 and xg < w and yg >= 0 and yg < h
      and pieces[i].map[yg * w + xg + 1] == 1
    then
      piece = i
      grabmode = "piece"
      grab_x = x
      grab_y = y
      return
    end
  end
end


function love.mousemoved(x, y)
  if won then return end
  if not grabmode then return end
  x, y = scr2map(x, y)
  local dx, dy = x - grab_x, y - grab_y

  if grabmode == "slider" then
    -- Check if it has room to slide
    local sldx = dx < 0 and math.ceil(sliders[slider].x)
                        or math.floor(sliders[slider].x)
    local sldy = sliders[slider].y
    local map1y = sldy - pieces[1].y
    local map2y = sldy - pieces[2].y
    -- If piece is at a non-integral position, we can't move
    if map1y ~= math.floor(map1y) --[[or map2y ~= math.floor(map2y)]] then
      return
    end

    local tgtx = dx < 0 and math.floor(sliders[slider].x + dx)
                      or math.ceil(sliders[slider].x + dx)
    if tgtx == sldx then return end -- no movement
    local last = sldx
    for i = sldx, tgtx, sgn(tgtx - sldx) do
      -- sliders[i].x references the slider in the first piece
      local mapx = i - pieces[1].x
      if mapx < 0 or mapx >= w or map1y < 0 or map1y >= h
        or pieces[1].map[map1y * w + mapx + 1] ~= 0
        or not (map2y >= h -- it can exit through the bottom - not invalid
                or (map2y >= 0
                    and pieces[2].map[map2y * w + mapx + 1] == 0))
      then
        -- can't move - set integral position closest to this
        sliders[slider].x = last
        grab_x = x
        return
      end
      last = i
    end
    -- Move possible - update slider position and grabbed position
    sliders[slider].x = sliders[slider].x + dx
    grab_x = x
    return
  end

  assert(grabmode == "piece")
  local closestvalue, closestindex
  for i = 1, 3 do
    local sldx = sliders[i].x
    -- If slider is at a non-integral position, we can't move
    if sldx ~= math.floor(sldx) then return end
    sldx = sldx - pieces[1].x
    assert(sldx >= 0 and sldx < w)
    local sldy = sliders[i].y - pieces[piece].y
    local tgty = sliders[i].y - (pieces[piece].y + dy)
    if dy < 0 then
      sldy = math.floor(sldy)
      tgty = math.ceil(tgty)
    else
      sldy = math.ceil(sldy)
      tgty = math.floor(tgty)
    end
    if tgty == sldy then return end -- no movement
    local last = sldy
    for j = sldy, tgty, sgn(tgty - sldy) do
      if j < 0 or j < h and pieces[piece].map[j * w + sldx + 1] ~= 0 then
        if dy < 0 then
          closestvalue = closestvalue and math.max(closestvalue, sliders[i].y - last)
            or sliders[i].y - last
        else
          closestvalue = closestvalue and math.min(closestvalue, sliders[i].y - last)
            or sliders[i].y - last
        end
        break
      end
      last = j
    end
  end
  if closestvalue then
    -- Collision detected in one of the sliders
    pieces[piece].y = closestvalue
  else
    pieces[piece].y = pieces[piece].y + dy
  end
  grab_y = y
end


function love.mousereleased(x, y, b)
  if won then return end
  if b ~= 1 or not grabmode then return end
  portland.forceOrient(false)

  if grabmode == "piece" then
    for i = 1, 2 do
      local y = pieces[i].y
      if math.abs(math.floor(y + 0.5) - y) < 0.4 then
        pieces[i].y = math.floor(y + 0.5)
      end
    end
  elseif grabmode == "slider" then
    for i = 1, 3 do
      local x = sliders[i].x
      if math.abs(math.floor(x + 0.5) - x) < 0.4 then
        sliders[i].x = math.floor(x + 0.5)
      end
    end
  end

  grabmode = false
end
--[[debug:
FIXME=true
MAP2SCR=map2scr
SCR2MAP=scr2map
PIECES=pieces
GRABMODE=grabmode
--]]
